
/* this file is a part of gau software, (C) Hwang chi-deok 1997, 1998         */

#include <string.h>		/* memmove */
#include <ctype.h>		/* isspace isalnum */
#include <gdk/gdkkeysyms.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkvscrollbar.h>
#include <gtk/gtkdrawingarea.h>
#include <gdk/hangulinput.h>
#include "gauchat.h"

#define LINE_BUFFER(y) (chat->buf[y])
#define CURRENT_LINE_BUFFER LINE_BUFFER(chat->current_y)
#define CURRENT_CHAR_BUFFER   ((CURRENT_LINE_BUFFER + chat->current_x))

enum {
    SEND,
    LINE_CHANGED,
    LAST_SIGNAL
};

typedef void (*GauChatFunction) (GauChat * chat);
typedef void (*GauChatSignal1) (GtkObject * object,
				gpointer arg1,
				gint arg2,
				gpointer data);

static void gau_chat_marshal_signal_1(GtkObject * object,
				      GtkSignalFunc func,
				      gpointer func_data,
				      GtkArg * arg);

#define CHAT_MASK ( GDK_EXPOSURE_MASK | \
					GDK_BUTTON_PRESS_MASK | \
					GDK_KEY_PRESS_MASK)

static void gau_chat_class_init(GauChatClass * class);
static void gau_chat_init(GauChat * chat);
static void gau_chat_destroy(GtkObject * object);

static gint gau_chat_expose(GtkWidget * w, GdkEventExpose * ev);
static gint gau_chat_button_press(GtkWidget * w, GdkEventButton * ev);
static gint gau_chat_key_press(GtkWidget * w, GdkEventKey * ev);
static void gau_chat_text_area_realize_callback(GtkWidget * w);
static void gau_chat_delete_forward(GauChat * chat);
static void gau_chat_delete_backward(GauChat * chat);
static void gau_chat_delete_forward_word(GauChat * chat);
static void gau_chat_delete_backward_word(GauChat * chat);
static void gau_chat_delete_eol(GauChat * chat);
static void gau_chat_delete_line(GauChat * chat);
static void gau_chat_expose_eol(GauChat * chat);
static void gau_chat_clear_line(GauChat * chat, int y, int x1, int x2);
static void gau_chat_show_cursor(GauChat * chat);
static void gau_chat_hide_cursor(GauChat * chat);
static void gau_chat_draw_line(GauChat * chat, int y, int x1, int x2);
static void gau_chat_move_forward(GauChat * chat);
static void gau_chat_move_backward(GauChat * chat);
static void gau_chat_move_upward(GauChat * chat);
static void gau_chat_move_downward(GauChat * chat);
static void gau_chat_move_eol(GauChat * chat);
static void gau_chat_move_bol(GauChat * chat);
static void gau_chat_move_forward_word(GauChat * chat);
static void gau_chat_move_backward_word(GauChat * chat);
static void gau_chat_move_scroll(GauChat * chat);
static void gau_chat_hangul_correct(GauChat * chat, char *buf);
static void gau_chat_hangul_insert(GauChat * chat, char *buf);
static gint gau_chat_check_offset(GauChat * chat);

static GtkDrawingAreaClass *parent_class = NULL;
static gint chat_signals[LAST_SIGNAL] =
{0};
static GauChatFunction control_keys[26] =
{
    gau_chat_move_bol,		/* a */
    gau_chat_move_backward,	/* b */
    NULL,			/* c */
    gau_chat_delete_forward,	/* d */
    gau_chat_move_eol,		/* e */
    gau_chat_move_forward,	/* f */
    NULL,			/* g */
    NULL,			/* h */
    NULL,			/* i */
    NULL,			/* j */
    gau_chat_delete_eol,	/* k */
    NULL,			/* l */
    NULL,			/* m */
    gau_chat_move_downward,	/* n */
    NULL,			/* o */
    gau_chat_move_upward,	/* p */
    NULL,			/* q */
    NULL,			/* r */
    NULL,			/* s */
    NULL,			/* t */
    gau_chat_delete_line,	/* u */
    NULL,			/* v */
    gau_chat_delete_backward_word,	/* w */
    NULL,			/* x */
    NULL,			/* y */
    NULL,			/* z */
};

static GauChatFunction alt_keys[26] =
{
    NULL,			/* a */
    gau_chat_move_backward_word,	/* b */
    NULL,			/* c */
    gau_chat_delete_forward_word,	/* d */
    NULL,			/* e */
    gau_chat_move_forward_word,	/* f */
    NULL,			/* g */
    NULL,			/* h */
    NULL,			/* i */
    NULL,			/* j */
    NULL,			/* k */
    NULL,			/* l */
    NULL,			/* m */
    NULL,			/* n */
    NULL,			/* o */
    NULL,			/* p */
    NULL,			/* q */
    NULL,			/* r */
    NULL,			/* s */
    NULL,			/* t */
    NULL,			/* u */
    NULL,			/* v */
    NULL,			/* w */
    NULL,			/* x */
    NULL,			/* y */
    NULL,			/* z */
};

guint
gau_chat_get_type(void)
{
    static guint chat_type = 0;
    if (!chat_type) {
	GtkTypeInfo chat_info =
	{
	    "GauChat",
	    sizeof(GauChat),
	    sizeof(GauChatClass),
	    (GtkClassInitFunc) gau_chat_class_init,
	    (GtkObjectInitFunc) gau_chat_init,
	    (GtkArgSetFunc) NULL,
	    (GtkArgGetFunc) NULL,
	};
	chat_type = gtk_type_unique(gtk_drawing_area_get_type(), &chat_info);
    }
    return chat_type;
}

static void gau_chat_class_init(GauChatClass * class)
{
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;

    object_class = (GtkObjectClass *) class;
    widget_class = (GtkWidgetClass *) class;
    parent_class = gtk_type_class(gtk_drawing_area_get_type());

    chat_signals[0] = gtk_signal_new("send",
				     GTK_RUN_LAST,
				     object_class->type,
				   GTK_SIGNAL_OFFSET(GauChatClass, send),
			     gau_chat_marshal_signal_1, GTK_TYPE_NONE, 2,
				     GTK_TYPE_POINTER, GTK_TYPE_INT);
    chat_signals[1] = gtk_signal_new("line_changed",
				     GTK_RUN_LAST,
				     object_class->type,
			   GTK_SIGNAL_OFFSET(GauChatClass, line_changed),
			gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
    gtk_object_class_add_signals(object_class, chat_signals, LAST_SIGNAL);
    widget_class->expose_event = gau_chat_expose;
    widget_class->key_press_event = gau_chat_key_press;
    widget_class->button_press_event = gau_chat_button_press;
    object_class->destroy = gau_chat_destroy;
}

static void gau_chat_init(GauChat * chat)
{
    int i;
    chat->offset = 0;
    chat->add = -1;
    chat->prefix1 = g_strdup("");
    chat->prefix2 = g_strdup("");
    chat->suffix = g_strdup("");
    chat->current_x = chat->current_y = 0;
    chat->has_temp_hangul = FALSE;
    for (i = 0; i < 20; i++)
	chat->fbinding[i] = NULL;
}

GtkWidget *
 gau_chat_new(GdkFont * font, int font_height, int font_ascent, int col, int row)
{
    GauChat *chat;
    gint mask;
    GtkWidget *w;
    int i;

    chat = gtk_type_new(gau_chat_get_type());
    w = GTK_WIDGET(chat);
    chat->buf = g_new0(char *, row);
    chat->buf_alloc = g_new(int, row);
    for (i = 0; i < row; i++) {
	chat->buf_alloc[i] = 50;
	chat->buf[i] = g_new(char, chat->buf_alloc[i]);
	chat->buf[i][0] = 0;
    }
    chat->col = col;
    chat->row = row;
    GTK_WIDGET_SET_FLAGS(chat, GTK_CAN_FOCUS);
    GTK_WIDGET_SET_FLAGS(chat, GTK_CAN_DEFAULT);
    mask = gtk_widget_get_events(w);
    gtk_widget_set_events(w, mask | CHAT_MASK | GDK_STRUCTURE_MASK);
    gtk_signal_connect(GTK_OBJECT(chat), "realize",
	     GTK_SIGNAL_FUNC(gau_chat_text_area_realize_callback), NULL);
    chat->font_width = gdk_string_width(font, " ");
    if (font_height <= 0)
	chat->font_height = font->ascent + font->descent;
    else
	chat->font_height = font_height;
    if (font_ascent <= 0)
	chat->font_ascent = font->ascent;
    else
	chat->font_ascent = font_ascent;
    chat->font = font;
    gtk_drawing_area_size(GTK_DRAWING_AREA(chat),
			col * chat->font_width, row * chat->font_height);
    return GTK_WIDGET(chat);
}

void gau_chat_binding_set(GauChat * chat, int index, char *buf)
{
    chat->fbinding[index] = buf;
}

void gau_chat_insert_text(GauChat * chat, char *buf, int len)
{
    while (len > 0) {
	if (*buf & 0x80) {
	    gau_chat_hangul_insert(chat, buf);
	    len -= 2;
	    buf += 2;
	} else {
	    gau_chat_insert_char(chat, *buf);
	    buf++;
	    len--;
	}
    }
    /*
       gau_chat_check_offset(chat);
     */
}

void gau_chat_insert_char(GauChat * chat, char c)
{
    char *s;
    static char *text = NULL;
    int len;
    /*
       g_print("%s: %d %d\n", __FUNCTION__, chat->current_x, chat->current_y);
     */
    if (c == '\r') {
	if (text)
	    g_free(text);
	if (chat->add > 0 && *CURRENT_LINE_BUFFER != '/') {
	    if (chat->add == 1)
		s = chat->prefix1;
	    else
		s = chat->prefix2;
	    text = g_strconcat (s, CURRENT_LINE_BUFFER, chat->suffix, NULL);
	} else {
	    text = g_strdup(CURRENT_LINE_BUFFER);
	}
	if (chat->current_y == chat->row - 1)
	    gau_chat_move_scroll(chat);
	else {
	    gau_chat_hide_cursor(chat);
	    chat->current_x = 0;
	    gau_chat_move_downward(chat);
	}
	gtk_signal_emit(GTK_OBJECT(chat), chat_signals[SEND],
			text, strlen(text));
	return;
    }
    s = CURRENT_CHAR_BUFFER;
    len = strlen(s);
    if ((len + chat->current_x) >= (chat->buf_alloc[chat->current_y] - 1)) {
	chat->buf_alloc[chat->current_y] += 50;
	chat->buf[chat->current_y] =
	    g_realloc(chat->buf[chat->current_y],
		      chat->buf_alloc[chat->current_y]);
	s = CURRENT_CHAR_BUFFER;
    }
    gau_chat_clear_line(chat, chat->current_y, chat->current_x, chat->current_x + len + 1);
    memmove(s + 1, s, len + 1);
    *s = c;
    /*
       g_print("%s\n", s);
     */
    gau_chat_expose_eol(chat);
    chat->current_x++;
    /*
       gau_chat_check_offset(chat);
     */
    gau_chat_show_cursor(chat);
}

void gau_chat_add_set(GauChat * chat, gint flag)
{
    chat->add = flag;
}

void gau_chat_prefix_set(GauChat * chat, char *s, gint flag)
{
    g_return_if_fail(s != NULL);
    g_return_if_fail(chat != NULL);
    g_return_if_fail(GAU_IS_CHAT(chat));

    if (flag == 1) {
	g_free(chat->prefix1);
	chat->prefix1 = g_strdup(s);
    } else {
	g_free(chat->prefix2);
	chat->prefix2 = g_strdup(s);
    }
}

void gau_chat_suffix_set(GauChat * chat, char *s)
{
    g_return_if_fail(s != NULL);
    g_return_if_fail(chat != NULL);
    g_return_if_fail(GAU_IS_CHAT(chat));

    g_free(chat->suffix);
    chat->suffix = g_strdup(s);
}


static void gau_chat_hangul_insert(GauChat * chat, char *buf)
{
    char *s;
    int len;
    gau_chat_hide_cursor(chat);
    s = CURRENT_CHAR_BUFFER;
    len = strlen(s);
    if ((len + chat->current_x) >= (chat->buf_alloc[chat->current_y] - 2)) {
	chat->buf_alloc[chat->current_y] += 50;
	chat->buf[chat->current_y] =
	    g_realloc(chat->buf[chat->current_y],
		      chat->buf_alloc[chat->current_y]);
	s = CURRENT_CHAR_BUFFER;
    }
    gau_chat_clear_line(chat, chat->current_y,
			chat->current_x, chat->current_x + len);
    memmove(s + 2, s, len + 1);
    *s = *buf;
    *(s + 1) = *(buf + 1);
    gau_chat_expose_eol(chat);
    chat->current_x += 2;
    /*
       gau_chat_check_offset(chat);
     */
    gau_chat_show_cursor(chat);
}

static void gau_chat_hangul_correct(GauChat * chat, char *buf)
{
    char *s = CURRENT_CHAR_BUFFER;
    *(s - 2) = *buf;
    *(s - 1) = *(buf + 1);
    gau_chat_clear_line(chat, chat->current_y,
			chat->current_x - 2,
			chat->current_x);
    gau_chat_draw_line(chat, chat->current_y,
		       chat->current_x - 2,
		       chat->current_x);
}
static int length_of_line(GauChat * chat, int y)
{
    return strlen(LINE_BUFFER(y));
}

static void gau_chat_show_cursor(GauChat * chat)
{
    int x, y;
    GdkGC *gc;
    /*
       g_print("%d %d %d\n", chat->current_x, chat->current_y, chat->offset);
     */
    if (gau_chat_check_offset(chat))
	return;
    x = chat->font_width * (chat->current_x - chat->offset);
    y = chat->font_height * chat->current_y;
    if (GTK_WIDGET_HAS_FOCUS(chat)) {
	gc = GTK_WIDGET(chat)->style->fg_gc[GTK_STATE_NORMAL];
    } else
	gc = GTK_WIDGET(chat)->style->black_gc;
    gdk_draw_line(GTK_WIDGET(chat)->window, gc,
		  x, y, x, y + chat->font_height - 1);
    if (chat->has_temp_hangul) {
	gdk_draw_line(GTK_WIDGET(chat)->window, gc,
		      x - chat->font_width * 2 + 1, 
		      y + chat->font_height - 1, 
		      x - 2, 
		      y + chat->font_height - 1);
    }
}

static void gau_chat_hide_cursor(GauChat * chat)
{
    int mul;
    /*
       g_print("%s %d %d\n", __FUNCTION__, chat->current_x, chat->current_y);
     */
    if (*CURRENT_CHAR_BUFFER & 0x80)
	mul = 2;
    else
	mul = 1;
    gdk_window_clear_area(GTK_WIDGET(chat)->window,
		     (chat->current_x - chat->offset) * chat->font_width,
			  chat->current_y * chat->font_height,
			  mul * chat->font_width, chat->font_height);
    if (*CURRENT_CHAR_BUFFER)
	gau_chat_draw_line(chat, chat->current_y,
			   chat->current_x, chat->current_x + mul);

}

static void gau_chat_move_bol(GauChat * chat)
{
    if (chat->current_x == 0)
	return;
    gau_chat_hide_cursor(chat);
    chat->current_x = 0;
    /*
       gau_chat_check_offset(chat);
     */
    gau_chat_show_cursor(chat);
}

static void gau_chat_move_eol(GauChat * chat)
{
    gau_chat_hide_cursor(chat);
    chat->current_x = length_of_line(chat, chat->current_y);
    /*
       gau_chat_check_offset(chat);
     */
    gau_chat_show_cursor(chat);
}


static void gau_chat_delete_forward(GauChat * chat)
{
    char *s = CURRENT_CHAR_BUFFER;
    int delta = 1;
    gau_chat_clear_line(chat, chat->current_y, chat->current_x,
			chat->current_x + strlen(s));
    if (*s & 0x80)
	delta = 2;
    memmove(s, s + delta, strlen(s + delta) + 1);
    gau_chat_expose_eol(chat);
    gau_chat_show_cursor(chat);
}

static void gau_chat_delete_backward(GauChat * chat)
{
    if (chat->current_x == 0) {
	/* already beginning of line */
	return;
    }
    gau_chat_move_backward(chat);
    gau_chat_delete_forward(chat);
}

static void gau_chat_delete_forward_word(GauChat * chat)
{
    guchar *s = CURRENT_CHAR_BUFFER;
    guchar *e = s;
    gau_chat_clear_line(chat, chat->current_y, chat->current_x,
			chat->current_x + strlen(s));
    if (isspace(*e)) {
	while (isspace(*e))
	    e++;
    }
    while (!isspace(*e) && *e)
	e++;
    memmove(s, e, strlen(e) + 1);
    gau_chat_expose_eol(chat);
    gau_chat_show_cursor(chat);
}

static void gau_chat_delete_backward_word(GauChat * chat)
{
    guchar *s = CURRENT_CHAR_BUFFER;
    guchar *e = s;
    if (chat->current_x == 0)
	return;
    gau_chat_clear_line(chat, chat->current_y, chat->current_x,
			chat->current_x + strlen(s));
    if (*s == 0) {
	s--;
	chat->current_x--;
    }
    if (isspace(*s)) {
	while (isspace(*s) && chat->current_x >= 0) {
	    chat->current_x--;
	    s--;
	}
    }
    while (!isspace(*s) && chat->current_x >= 0) {
	chat->current_x--;
	s--;
    }
    s++;
    chat->current_x++;
    gau_chat_clear_line(chat, chat->current_y, chat->current_x,
			chat->current_x + s - e);
    memmove(s, e, strlen(e) + 1);
    gau_chat_expose_eol(chat);
    gau_chat_show_cursor(chat);
}

static void gau_chat_delete_line(GauChat * chat)
{
    gau_chat_clear_line(chat, chat->current_y, 0, strlen(CURRENT_LINE_BUFFER));
    chat->current_x = 0;
    *CURRENT_LINE_BUFFER = 0;
    gau_chat_show_cursor(chat);
}


static void gau_chat_delete_eol(GauChat * chat)
{
    gau_chat_clear_line(chat, chat->current_y, chat->current_x,
			strlen(CURRENT_LINE_BUFFER + chat->current_x));
    *CURRENT_CHAR_BUFFER = 0;
    gau_chat_show_cursor(chat);
}


/* right key action */
static void gau_chat_move_forward(GauChat * chat)
{
    if (chat->current_x == length_of_line(chat, chat->current_y)) {
	gau_chat_move_bol(chat);
	gau_chat_move_downward(chat);
	return;
    }
    gau_chat_hide_cursor(chat);
    chat->current_x++;
    if (is_second_han(CURRENT_LINE_BUFFER, chat->current_x))
	chat->current_x++;
    /*
       gau_chat_check_offset(chat);
     */
    gau_chat_show_cursor(chat);
}

static void gau_chat_move_forward_word(GauChat * chat)
{
    guchar *s;
    if (chat->current_x == length_of_line(chat, chat->current_y)) {
	gau_chat_move_forward(chat);
	return;
    }
    s = CURRENT_CHAR_BUFFER;
    gau_chat_hide_cursor(chat);
    while (*s & 0x80 || isalnum(*s) || *s == '_') {
	s++;
	chat->current_x++;
    }
    if (isspace(*s)) {
	do {
	    s++;
	    chat->current_x++;
	} while (isspace(*s));
    }
    gau_chat_show_cursor(chat);
}


/* left key action */
static void gau_chat_move_backward(GauChat * chat)
{
    if (chat->current_x == 0) {
	if (chat->current_y == 0) {
	    gdk_beep();
	    return;
	}
	gau_chat_move_upward(chat);
	gau_chat_move_eol(chat);
	return;
    }
    gau_chat_hide_cursor(chat);
    chat->current_x--;
    if (is_second_han(CURRENT_LINE_BUFFER, chat->current_x))
	chat->current_x--;
    /*
       gau_chat_check_offset(chat);
     */
    gau_chat_show_cursor(chat);
}

static void gau_chat_move_backward_word(GauChat * chat)
{
    guchar *s = CURRENT_CHAR_BUFFER;
    /*
       g_print("%s\n", __FUNCTION__);
     */
    if (chat->current_x == 0) {
	gau_chat_move_backward(chat);
	return;
    }
    /*
       g_print("%d \n", chat->current_x);
     */
    gau_chat_hide_cursor(chat);
    s--;
    chat->current_x--;
    if (isspace(*s))
	do {
	    s--;
	    chat->current_x--;
	} while (isspace(*s) && chat->current_x >= 0);
    while (isalnum(*s) || *s == '_' || (*s & 0x80)) {
	if (chat->current_x == 0)
	    break;
	s--;
	chat->current_x--;
    };
    if (chat->current_x != 0)
	chat->current_x++;
    /*
       g_print("%d \n", chat->current_x);
     */
    gau_chat_show_cursor(chat);
}


/* up key action */
static void gau_chat_move_upward(GauChat * chat)
{
    int len;
    if (chat->current_y == 0) {	/* first line: */
	return;
    } else {
	gau_chat_hide_cursor(chat);
	chat->current_y--;
	gtk_signal_emit(GTK_OBJECT(chat), chat_signals[LINE_CHANGED]);
    }
    len = length_of_line(chat, chat->current_y);
    if (len < chat->current_x)
	chat->current_x = len;
    if (is_second_han(CURRENT_LINE_BUFFER, chat->current_x))
	chat->current_x++;
    /*
       gau_chat_check_offset(chat);
     */
    gau_chat_show_cursor(chat);
}

/* down key action */
static void gau_chat_move_downward(GauChat * chat)
{
    int len;
    if (chat->current_y == chat->row - 1) {	/* last line: . */
	return;
    } else {
	gau_chat_hide_cursor(chat);
	chat->current_y++;
	gtk_signal_emit(GTK_OBJECT(chat), chat_signals[LINE_CHANGED]);
    }
    len = length_of_line(chat, chat->current_y);
    if (len < chat->current_x)
	chat->current_x = len;
    if (is_second_han(CURRENT_LINE_BUFFER, chat->current_x))
	chat->current_x++;
    /*
       gau_chat_check_offset(chat);
     */
    gau_chat_show_cursor(chat);
}

static void gau_chat_move_scroll(GauChat * chat)
{
    int y;
    char *buf0 = chat->buf[0];
    int buf_alloc0 = chat->buf_alloc[0];
    gdk_window_clear(GTK_WIDGET(chat)->window);

    for (y = 0; y < chat->row - 1; y++) {
	chat->buf[y] = chat->buf[y + 1];
	chat->buf_alloc[y] = chat->buf_alloc[y + 1];
    }
    chat->buf[chat->row - 1] = buf0;
    chat->buf_alloc[chat->row - 1] = buf_alloc0;
    buf0[0] = 0;
    chat->current_x = 0;
    gtk_widget_draw(GTK_WIDGET(chat), NULL);
    gau_chat_show_cursor(chat);
}



static void gau_chat_expose_eol(GauChat * chat)
{
    int x1 = chat->current_x;
    int x2;
    char *buf = CURRENT_CHAR_BUFFER;
    if (*buf == 0)
	return;
    x2 = x1 + strlen(buf);
    gau_chat_draw_line(chat, chat->current_y, x1, x2);
}

static void gau_chat_draw_line(GauChat * chat, int y, int x1, int x2)
{
    char *buf = LINE_BUFFER(y);
    if (x1 < chat->offset)
	x1 = chat->offset;
    if (x2 <= x1)
	return;
    if (is_second_han(buf, x1))
	x1--;
    if (is_first_han(buf, x2 - 1))
	x2++;
    gdk_draw_text(GTK_WIDGET(chat)->window, chat->font, chat->gc,
		  (x1 - chat->offset) * chat->font_width,
		  y * chat->font_height + chat->font_ascent,
		  buf + x1, x2 - x1);
}

static void gau_chat_hide_underline (GauChat * chat)
{
    if (gau_chat_check_offset(chat))
	return;
    gau_chat_clear_line (chat, chat->current_y, 
    			 chat->current_x - 2, chat->current_x);
    gau_chat_draw_line (chat, chat->current_y, 
                        chat->current_x - 2, 
			chat->current_x);
}


static void gau_chat_clear_line(GauChat * chat, int y, int x1, int x2)
{
    if (x1 < chat->offset)
	x1 = chat->offset;
    if (x2 <= x1)
	return;
    gdk_window_clear_area(GTK_WIDGET(chat)->window,
			  (x1 - chat->offset) * chat->font_width,
			  y * chat->font_height,
			  (x2 - x1) * chat->font_width,
			  chat->font_height);
}


static gint
 gau_chat_key_press(GtkWidget * w, GdkEventKey * ev)
{
    gint key = ev->keyval;
    int n;
    char buf[30];
    GauChat *chat = GAU_CHAT(w);
    if (key >= GDK_Shift_L && key <= GDK_Alt_R)
	return FALSE;
    if (hangul_input_state_get()) {
	if ((key < 0x20 || key >= 0xff) && key != GDK_BackSpace) {
	    hangul_input_clear();
	    goto not_han;
	}
	if (key == GDK_space && ev->state & GDK_SHIFT_MASK) {
	    hangul_input_clear();
	    if (chat->has_temp_hangul) 
	       gau_chat_hide_underline (chat);
	    chat->has_temp_hangul = FALSE;
	    hangul_input_state_set(FALSE);
	    return TRUE;
	}
	if (ev->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK)) {
	    hangul_input_clear();
	    goto not_han;
	}
	n = hangul_input_process(buf, &key);
	switch (n) {
	case AUTOMATA_SPECIAL:
	case AUTOMATA_NULL:
	    goto not_han;
	case AUTOMATA_NEW:
	    if (chat->has_temp_hangul)
	        gau_chat_hide_underline (chat);
	    else
	        chat->has_temp_hangul = TRUE;
	    gau_chat_hangul_insert(chat, buf);
	    break;
	case AUTOMATA_CORRECT:
	    gau_chat_hangul_correct(chat, buf);
	    gau_chat_show_cursor (chat);
	    break;
	case AUTOMATA_CORRECT_NEW:
	    gau_chat_hangul_correct(chat, buf);
	    gau_chat_hide_underline (chat);
	    gau_chat_hangul_insert(chat, buf + 2);
	    break;
	case AUTOMATA_ERROR:
	    gdk_beep();
	    break;
	default:
	    g_warning("in automata\n");
	}
	return TRUE;
    }
  not_han:
    if (chat->has_temp_hangul) 
       gau_chat_hide_underline (chat);
    chat->has_temp_hangul = FALSE;
    if (key == GDK_space && ev->state & GDK_SHIFT_MASK) {
	hangul_input_state_set(TRUE);
	return TRUE;
    }
    if (key >= GDK_F1 && key <= GDK_F12) {
	key = key - GDK_F1;
	if (ev->state & GDK_SHIFT_MASK && key < 10) {
	    key += 10;
	}
	if (chat->fbinding[key]) {
	    gau_chat_insert_text(chat, chat->fbinding[key],
				 strlen(chat->fbinding[key]));
	    return TRUE;
	}
	return FALSE;
    }
    if (ev->state & GDK_CONTROL_MASK) {
	if ((key >= 'A') && (key <= 'Z'))
	    key -= 'A' - 'a';
	if ((key >= 'a') && (key <= 'z') && control_keys[key - 'a']) {
	    (*control_keys[key - 'a']) (chat);
	    return TRUE;
	}
    } else if (ev->state & GDK_MOD1_MASK) {
	if ((key >= 'A') && (key <= 'Z'))
	    key -= 'A' - 'a';
	if ((key >= 'a') && (key <= 'z') && alt_keys[key - 'a']) {
	    (*alt_keys[key - 'a']) (chat);
	    return TRUE;
	}
    } else {
	if (key >= GDK_KP_0 && key <= GDK_KP_9) {
	    key = key - GDK_KP_0 + '0';
	} else
	    switch (key) {
	    case GDK_KP_Divide:
		key = '/';
		break;
	    case GDK_KP_Multiply:
		key = '*';
		break;
	    case GDK_KP_Subtract:
		key = '-';
		break;
	    case GDK_KP_Add:
		key = '+';
		break;
	    default:
		break;
	    }
	if (key >= 0x20 && key <= 0xff) {
	    /* normal */
	    gau_chat_insert_char(chat, key);
	    return TRUE;
	}
	switch (key) {
	case GDK_Return:
	case GDK_KP_Enter:
	    gau_chat_insert_char(chat, '\r');
	    return TRUE;
	case GDK_Left:
	case GDK_KP_Left:
	    gau_chat_move_backward(chat);
	    return TRUE;
	case GDK_Right:
	case GDK_KP_Right:
	    gau_chat_move_forward(chat);
	    return TRUE;
	case GDK_Up:
	case GDK_KP_Up:
	    gau_chat_move_upward(chat);
	    return TRUE;
	case GDK_Down:
	case GDK_KP_Down:
	    gau_chat_move_downward(chat);
	    return TRUE;
	case GDK_Delete:
	case GDK_KP_Delete:
	    gau_chat_delete_forward(chat);
	    return TRUE;
	case GDK_BackSpace:
	    gau_chat_delete_backward(chat);
	    return TRUE;
	}
    }
    return FALSE;
}

static gint
 gau_chat_expose(GtkWidget * w, GdkEventExpose * ev)
{
    int x1, x2;
    int y1, y2;
    int i1, i2;
    int y;
    int len;
    GauChat *chat;
    chat = GAU_CHAT(w);
    x1 = ev->area.x / chat->font_width;
    x2 = (ev->area.x + ev->area.width) / chat->font_width;
    y1 = ev->area.y / chat->font_height;
    y2 = (ev->area.y + ev->area.height - 1) / chat->font_height;
    for (y = y1; y <= y2; y++) {
	len = length_of_line(chat, y);
	if (x1 + chat->offset >= len)
	    continue;
	if (is_second_han(LINE_BUFFER(y), x1 + chat->offset))
	    i1 = x1 - 1;
	else
	    i1 = x1;
	if (x2 + chat->offset > len)
	    i2 = len - chat->offset;
	else {
	    if (is_first_han(LINE_BUFFER(y), x2 + chat->offset))
		i2 = x2 + 1;
	    else
		i2 = x2;
	}
	gau_chat_draw_line(chat, y, i1 + chat->offset, i2 + chat->offset + 1);
	if (chat->current_y == y) {
	    if (chat->current_x <= i2 + chat->offset
		&& chat->current_x >= i1 + chat->offset)
		gau_chat_show_cursor(chat);
	}
    }
    return FALSE;
}

static gint
 gau_chat_button_press(GtkWidget * w, GdkEventButton * ev)
{
    int len;
    GauChat *chat;
    int x, y;
    if (!GTK_WIDGET_HAS_FOCUS(w))
	gtk_widget_grab_focus(w);
    chat = GAU_CHAT(w);
    x = ev->x / chat->font_width;
    y = ev->y / chat->font_height;
    gau_chat_hide_cursor(chat);
    if (x > (len = length_of_line(chat, y)))
	x = len;
    else if (is_second_han(LINE_BUFFER(y), x))
	x++;
    chat->current_x = x;
    chat->current_y = y;
    gau_chat_show_cursor(chat);
    return FALSE;
}

static gint
 gau_chat_check_offset(GauChat * chat)
{
    if (chat->current_x - chat->offset >= chat->col) {
	do {
	    chat->offset += chat->col / 2;
	} while (chat->current_x - chat->offset >= chat->col);
	gdk_window_clear(GTK_WIDGET(chat)->window);
	gtk_widget_draw(GTK_WIDGET(chat), NULL);
	return TRUE;
    } else if (chat->current_x < chat->offset) {
	do {
	    chat->offset -= chat->col / 2;
	} while (chat->current_x < chat->offset);
	gdk_window_clear(GTK_WIDGET(chat)->window);
	gtk_widget_draw(GTK_WIDGET(chat), NULL);
	return TRUE;
    }
    return FALSE;
}

static void gau_chat_text_area_realize_callback(GtkWidget * w)
{
    GtkStyle *style = gtk_widget_get_style(w);
    GauChat *chat = GAU_CHAT(w);
    gdk_window_set_background(w->window, &style->white);
    chat->gc = gdk_gc_new(w->window);
    gdk_gc_set_font(chat->gc, chat->font);
    gdk_gc_set_foreground(chat->gc, &style->black);
}


static void gau_chat_marshal_signal_1(GtkObject * object,
				      GtkSignalFunc func,
				      gpointer func_data,
				      GtkArg * args)
{
    GauChatSignal1 rfunc;
    rfunc = (GauChatSignal1) func;
    (*rfunc) (object, GTK_VALUE_POINTER(args[0]), GTK_VALUE_INT(args[1]), func_data);
}

static void gau_chat_destroy(GtkObject * object)
{
    GauChat *chat = GAU_CHAT(object);
    g_free(chat->buf);
    g_free(chat->prefix1);
    g_free(chat->prefix2);
    g_free(chat->suffix);
    if (GTK_OBJECT_CLASS(parent_class)->destroy)
	(*GTK_OBJECT_CLASS(parent_class)->destroy) (object);
}
