/* 

 * libxmb   by Hwang chideok(cdhwang@sr.hei.co.kr).
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include <X11/Xlibint.h>
#define WC_DEFINE 0

#if WC_DEFINE
#ifdef CI_GET_CHAR_INFO_1D
typedef wchar_t Wchar_t;
#define CI_NONEXISTCHAR(cs) (((cs)->width == 0) && \
			     (((cs)->rbearing|(cs)->lbearing| \
			       (cs)->ascent|(cs)->descent) == 0))

/* 
 * CI_GET_CHAR_INFO_1D - return the charinfo struct for the indicated 8bit
 * character.  If the character is in the column and exists, then return the
 * appropriate metrics (note that fonts with common per-character metrics will
 * return min_bounds).  If none of these hold true, try again with the default
 * char.
 */
#define CI_GET_CHAR_INFO_1D(fs,col,def,cs) \
{ \
    cs = def; \
    if (col >= fs->min_char_or_byte2 && col <= fs->max_char_or_byte2) { \
	if (fs->per_char == NULL) { \
	    cs = &fs->min_bounds; \
	} else { \
	    cs = &fs->per_char[(col - fs->min_char_or_byte2)]; \
	    if (CI_NONEXISTCHAR(cs)) cs = def; \
	} \
    } \
}

#define CI_GET_DEFAULT_INFO_1D(fs,cs) \
  CI_GET_CHAR_INFO_1D (fs, fs->default_char, NULL, cs)
#else
#define  xconst const 
#endif
#endif

int convert_ks_to_johabfont(const unsigned char *src, XChar2b * des, int len);
int convert_ks_to_johab844font(const unsigned char *src, XChar2b * des, int len);

#define DEBUG 0
#define _FONT_SET(fontset) ((_FontSet *)(fontset))
#define ENG_FONT(fontset) ((XFontStruct *)_FONT_SET(fontset)->fs[0])
#define H_FONT(fontset) ((XFontStruct *)_FONT_SET(fontset)->fs[1])
#define E_NAME(fontset) ((XFontStruct *)_FONT_SET(fontset)->name[0])
#define H_NAME(fontset) ((XFontStruct *)_FONT_SET(fontset)->name[1])
#define H_FONT_WIDTH(fontset) (H_FONT((fontset))->max_bounds.width)
#define ENG_DRAW(fontset) (*_FONT_SET(fontset)->eng.draw)
#define H_DRAW(fontset) (*_FONT_SET(fontset)->han.draw)

static int default_draw(Display * display, Drawable drawable, GC gc, int x, int y,
			_Xconst char *text, int len, XFontSet fontset);
static int ksc1_draw(Display * display, Drawable drawable, GC gc, int x, int y,
		     _Xconst char *text, int len, XFontSet fontset);
static int ksc0_draw(Display * display, Drawable drawable, GC gc, int x, int y,
		     _Xconst char *text, int len, XFontSet fontset);
static int johab_draw(Display * display, Drawable drawable, GC gc, int x, int y,
		      _Xconst char *text, int len, XFontSet fontset);
static int johab844_draw(Display * display, Drawable drawable, GC gc, int x, int y,
			 _Xconst char *text, int len, XFontSet fontset);
#if WC_DEFINE
static int wc2mb(_Xconst wchar_t * text, int len);
#endif

typedef struct __FontInfo _FontInfo;

typedef int (*DrawProc) (Display * display, Drawable d, GC gc, int x, int y, const char *str, int len, XFontSet fontset);
typedef _FontInfo *(*OpenProc) (Display * display, char *fontname);
typedef void (*CloseProc) (_FontInfo * font);
typedef int (*WidthProc) (_FontInfo * font, char *text, int len);
typedef void *_XFontStruct;
#if WC_DEFINE
static char *extent_buf = NULL;
static int extent_buf_size = 0;
#endif

struct __FontInfo {
    DrawProc draw;
    OpenProc open;
    CloseProc close;
    WidthProc width;
};

typedef struct {
    XFontStruct *fs[3];
    char **name;
    char *base_name;
    _FontInfo eng;
    _FontInfo han;
    XFontSetExtents extents;
    int hlift;
} _FontSet;

#if WC_DEFINE
void XwcDrawString(Display * display, Drawable window, XFontSet fontset, GC gc,
		   int x, int y, _Xconst wchar_t * text, int len)
{
    int n;
    n = wc2mb(text, len);
    XmbDrawString(display, window, fontset, gc, x, y, (_Xconst char *) extent_buf, n);
}
#endif

void XmbDrawString(Display * display, Drawable window, XFontSet fontset, GC gc,
		   int x, int y, _Xconst char *s, int len)
{
    const char *p;
#if DEBUG
    printf("drawstring\n");
    {
	int i = 0;
	putchar('\'');
	for (i = 0; i < len; i++)
	    putchar(s[i]);
	putchar('\'');
	putchar('\n');
    }
#endif

    while (len > 0) {
	if (*s & 0x80) {
	    for (p = s; len > 1 && (*s & 0x80) && (*(s + 1) & 0x80) ; s += 2, len -= 2);
	    x = H_DRAW(fontset) (display, window, gc, x, y - _FONT_SET(fontset)->hlift, p, ((int) (s - p)) / 2, fontset);
	} 
	if (len-- <= 0) break;
	for (p = s++; len > 0 && !(*s & 0x80); s++, len--);
	x = ENG_DRAW(fontset) (display, window, gc, x, y, p, s - p, fontset);
    }
}

#if WC_DEFINE
void XwcDrawImageString(Display * display, Drawable window, XFontSet fontset, GC gc,
			int x, int y, _Xconst wchar_t* text, int len)
{
    int n;
#if DEBUG
    printf("%s\n", __FUNCTION__);
#endif
    n = wc2mb(text, len);
    XmbDrawImageString(display, window, fontset, gc, x, y, (_Xconst char *) extent_buf, n);
}
#endif

void XmbDrawImageString(Display * display, Drawable window, XFontSet fontset, GC gc,
			int x, int y, _Xconst char *text, int len)
{
    XGCValues values;
    XFontSetExtents *extent;
    int width;
#if DEBUG
    printf("%s\n", __FUNCTION__);
#endif

    XGetGCValues(display, gc, (GCFunction | GCForeground | GCBackground | GCFillStyle), &values);

    XSetFunction(display, gc, GXcopy);
    XSetForeground(display, gc, values.background);
    XSetFillStyle(display, gc, FillSolid);

    extent = XExtentsOfFontSet(fontset);
    width = XmbTextEscapement(fontset, text, len);
    XFillRectangle(display, window, gc, x, y + extent->max_logical_extent.y, width, extent->max_logical_extent.height);
    XChangeGC(display, gc, (GCFunction | GCForeground | GCFillStyle), &values);

    XmbDrawString(display, window, fontset, gc, x, y, text, len);
}

char *
 get_fullname(Display * display, XFontStruct * fs)
{
    Atom atom_FONT;
    XFontProp *pp;
    char *fullname = NULL;
    int i;

    atom_FONT = XInternAtom(display, "FONT", False);
    pp = fs->properties;
    for (i = 0; i < fs->n_properties; i++) {
	if (pp->name == atom_FONT) {
	    fullname = XGetAtomName(display, pp->card32);
	    break;
	}
	pp++;
    }
    return fullname;
}


char **
 parse_fontname(Display * display, const char *font_name, int *n)
{
    char **name_list;
    char *eng_name = NULL, *han_name = NULL;
    char *fname;
    char *s;
    char **list;
    int count;
    int i;
    int len;
    if ((s = strchr(font_name, ',')) != NULL) {
	eng_name = malloc(s - font_name + 1);
	memcpy(eng_name, font_name, s - font_name + 1);
	eng_name[s - font_name] = 0;
	s++;
	while (isspace((unsigned char )*s)) s++;
	han_name = strdup(s);
	name_list = (char **) malloc(sizeof(char *) * 2);
	name_list[0] = eng_name;
	name_list[1] = han_name;
	*n = 2;
	return name_list;
    }
    list = XListFonts(display, font_name, 1024, &count);
    for (i = 0; i <= count; i++) {
	fname = list[i];
	len = strlen(fname);
	if (strcasecmp(fname + len - 9, "iso8859-1") == 0) {
	    eng_name = strdup(fname);
	    break;
	}
    }

    for (i = 0; i <= count; i++) {
	fname = list[i];
	len = strlen(fname);
	if (strncasecmp(fname + len - 14, "ksc5601.1987", 12) == 0) {
	    han_name = strdup(fname);
	    break;
	} else if (strcasecmp(fname + len - 7, "johab-1") == 0 ||
		   strcasecmp(fname + len - 8, "johabs-1") == 0 ||
		   strcasecmp(fname + len - 9, "johabsh-1") == 0) {
	    han_name = strdup(fname);
	    break;
	}
    }
    if (han_name == NULL)
	han_name = strdup("-daewoo-mincho-medium-r-normal--16-120-100-100-c-160-ksc5601.1987-0");
    name_list = (char **) malloc(sizeof(char) * 2);
    name_list[0] = eng_name;
    name_list[1] = han_name;
    XFreeFontNames(list);
    *n = 2;
    return name_list;
}

XFontSet
XCreateFontSet(Display * display, _Xconst char *font_name,
	       char ***missing_charset,
	       int *missing_set_count, char **def_string)
{
    char *hfont_name = NULL, *efont_name = NULL;
    char *hfont_fullname = NULL;
    char *xfont_fullname = NULL;
    _FontSet *fset;
    XFontStruct *xfont = NULL, *hfont = NULL;
    int ascent, descent;
    char *base_name;
    char **name_list;
    char *charset_registry;
    int count;

    *missing_set_count = 0;
    *def_string = "";



    fset = (_FontSet *) malloc(sizeof(_FontSet));

    base_name = strdup(font_name);
#if DEBUG
    printf("create font_name: %s\n", font_name);
#endif
    name_list = parse_fontname(display, font_name, &count);
    efont_name = name_list[0];
    if (count >= 2)
	hfont_name = name_list[1];
#if 0
    efont_name = strdup(font_name);
    if ((s = strchr(font_name, ',')) != NULL) {
	s = strchr(efont_name, ',');
	hfont_name = s + 1;
	while (isspace(*hfont_name))
	    hfont_name++;
	s--;
	while (isspace(*s))
	    s--;
	*(s + 1) = '\0';
    }
#endif

    xfont = XLoadQueryFont(display, efont_name);
    if (xfont) {
	xfont_fullname = get_fullname(display, xfont);
    } else {
	printf("fail to open english font:%s\n", efont_name);
    }
    _FONT_SET(fset)->eng.draw = default_draw;
    _FONT_SET(fset)->name = name_list;

    if (hfont_name)
	hfont = XLoadQueryFont(display, hfont_name);

    if (!hfont) {
	fprintf(stderr, "ѱ Ʈ %s  ϴ.\n", hfont_name);
	fprintf(stderr, "ȭ ϼż xlsfonts \n");
	fprintf(stderr, " ȯ Ʈ ֽʽÿ\n");
	exit(2);
    }
    ENG_FONT(fset) = xfont;
    H_FONT(fset) = hfont;
    _FONT_SET(fset)->fs[2] = NULL;
    _FONT_SET(fset)->base_name = base_name;

    charset_registry = NULL;
    {
	XFontProp *pp;
	int i;
	static char charset_buf[64];
	char *charset_encoding = NULL;
	Atom atom_FONT = XInternAtom(display, "FONT", False);
	Atom atom_CHARSET = XInternAtom(display, "CHARSET_REGISTRY", False);
	Atom atom_CHARSET_ENCODING = XInternAtom(display, "CHARSET_ENCODING", False);

	for (i = 0, pp = hfont->properties; i < hfont->n_properties; i++, pp++)
	    if (pp->name == atom_CHARSET) {
		charset_registry = XGetAtomName(display, pp->card32);
	    } else if (pp->name == atom_CHARSET_ENCODING) {
		charset_encoding = XGetAtomName(display, pp->card32);
	    } else if (pp->name == atom_FONT) {
		hfont_fullname = XGetAtomName(display, pp->card32);
	    }
	if (charset_registry && charset_encoding) {
	    sprintf(charset_buf, "%s-%s", charset_registry, charset_encoding);
	    charset_registry = charset_buf;
	}
    }

    if (charset_registry == NULL) {
	charset_registry = hfont_fullname;
    }
    if (strstr(charset_registry, "KSC5601.1987-0") != NULL ||
	strstr(charset_registry, "ksc5601.1987-0") != NULL) {
	char ga[3] = "";
	ga[0] &= 0x7F;
	ga[1] &= 0x7F;
	H_FONT_WIDTH(fset) = XTextWidth16(hfont, (XChar2b *) ga, 1);
	_FONT_SET(fset)->han.draw = ksc0_draw;
    } else if (strstr(charset_registry, "ksc5601.1987-1") != NULL ||
	       strstr(charset_registry, "KSC5601.1987-1") != NULL) {
	_FONT_SET(fset)->han.draw = ksc1_draw;
    } else if (strstr(charset_registry, "johab844") != NULL ||
	       strstr(charset_registry, "JOHAB844") != NULL) {
	_FONT_SET(fset)->han.draw = johab844_draw;
    } else if (strstr(charset_registry, "johab") != NULL ||
	       strstr(charset_registry, "JOHAB") != NULL) {
	_FONT_SET(fset)->han.draw = johab_draw;
    } else {
	char ga[3] = "";
	ga[0] &= 0x7F;
	ga[1] &= 0x7F;
	H_FONT_WIDTH(fset) = XTextWidth16(hfont, (XChar2b *) ga, 1);
	fprintf(stderr, "cannot resolve font encoding method of %s (%s)\n", name_list[1], charset_registry);
	_FONT_SET(fset)->han.draw = ksc0_draw;
    }


    if (!xfont)
	(*missing_set_count)++;
    if (!hfont)
	(*missing_set_count)++;

    // FIXME fset->extents;
    ascent = xfont->ascent;
    descent = xfont->descent;
    if (hfont) {
	if (hfont->descent < 0) {
	    hfont->descent = H_FONT_WIDTH(fset) - hfont->ascent;
	    _FONT_SET(fset)->hlift = hfont->descent - xfont->descent;
	    hfont->ascent += _FONT_SET(fset)->hlift;
	    hfont->descent -= _FONT_SET(fset)->hlift;
	    _FONT_SET(fset)->hlift = -1;
	} else {
	    _FONT_SET(fset)->hlift = hfont->descent - xfont->descent;
	    hfont->ascent += _FONT_SET(fset)->hlift;
	    hfont->descent -= _FONT_SET(fset)->hlift;
	}
	if (ascent < hfont->ascent)
	    ascent = hfont->ascent;
	if (descent < hfont->descent)
	    descent = hfont->descent;
    }
    /* FIXME */
    if (*missing_set_count) {
	char *str;
	str = malloc(5);
	strcpy(str, "C");
	strcpy(str + 2, "ko");
	*missing_charset = malloc(sizeof(char *) * 2);
	*missing_charset[0] = str;
	*missing_charset[1] = str + 2;
    }
    fset->extents.max_logical_extent.x = 0;
    fset->extents.max_logical_extent.y = -ascent;
    fset->extents.max_logical_extent.height = ascent + descent;
    /* FIXME: must check whether we succeed in loading hangul font */
    fset->extents.max_logical_extent.width = H_FONT_WIDTH(fset);
    /* FIXME: the following is wrong obviously */
    fset->extents.max_ink_extent.x = 0;
    fset->extents.max_ink_extent.y = -ascent;
    fset->extents.max_ink_extent.height = ascent + descent;
    fset->extents.max_ink_extent.width = H_FONT_WIDTH(fset);
#if DEBUG
    printf("name: %s\n", font_name);
    printf("ascent: %d, descent: %d\n", ascent, descent);
#endif
    return (XFontSet) fset;
}

void XFreeFontSet(Display * display, XFontSet fontset)
{
    XFreeFont(display, ENG_FONT(fontset));
    XFree(E_NAME(fontset));
    XFree(_FONT_SET(fontset)->base_name);
    XFree(_FONT_SET(fontset)->name[0]);
    if (_FONT_SET(fontset)->name[1])
	XFree(_FONT_SET(fontset)->name[1]);
    XFree(_FONT_SET(fontset)->name);
    if (H_FONT(fontset))
	XFreeFont(display, H_FONT(fontset));
    if (H_NAME(fontset))
	XFree(H_NAME(fontset));
    XFree(fontset);
}

int XFontsOfFontSet(XFontSet font_set, XFontStruct *** font_struct_list, char ***font_name_list)
{
    *font_struct_list = _FONT_SET(font_set)->fs;
    *font_name_list = _FONT_SET(font_set)->name;
    return 2;
}

Bool
XContextualDrawing(XFontSet font_set)
{
    /*
       printf("Warning:%s in %s Not Implemented\n", __FUNCTION__, __FILE__);
     */
    return False;
}

Bool
XDirectionalDependentDrawing(XFontSet font_set)
{
    /*
       printf("Warning:%s in %s Not Implemented\n", __FUNCTION__, __FILE__);
     */
    return False;
}

char *
 XBaseFontNameListOfFontSet(XFontSet font_set)
{
    return _FONT_SET(font_set)->base_name;
}


char *XLocaleOfFontSet(XFontSet font_set)
{
    /*
       printf("Warning:%s in %s Not Implemented\n", __FUNCTION__, __FILE__);
     */
    return "ko";
}

XFontSetExtents *
 XExtentsOfFontSet(XFontSet fontset)
{
    return &(_FONT_SET(fontset)->extents);
}

#if WC_DEFINE
int XwcTextEscapement(XFontSet fontset, _Xconst wchar_t * text, int len)
{
    int n;
#if DEBUG
    printf("%s\n", __FUNCTION__);
#endif
    n = wc2mb(text, len);
    return XmbTextEscapement(fontset, (_Xconst char *) extent_buf, n);
}
#endif

int XmbTextEscapement(XFontSet fontset, _Xconst char *s, int len)
{
    XFontStruct *xfont, *hfont;
    int hfont_width_over_two;
    const char *p;
    int width;

    width = 0;
    xfont = ENG_FONT(fontset);
    hfont = H_FONT(fontset);

    if (hfont == NULL) {
	return XTextWidth(xfont, s, len);
    }
    hfont_width_over_two = H_FONT_WIDTH(fontset) / 2;

    while (len > 0) {
	if (*s & 0x80) {
	    for (p = s; (len > 0) && (*s & 0x80); s += 2, len -= 2);
	    if (len < 0)
		s--;
	    width += hfont_width_over_two * (s - p);
	} else {
	    for (p = s; (len > 0) && !(*s & 0x80); s++, len--);
	    width += XTextWidth(xfont, p, (s - p));
	}
    }
    return width;
}

#if WC_DEFINE
/* hmm.. locale supports in Linux is really poor.
   this is the quick hack about function wcstombs */
int wc2mb(_Xconst wchar_t * text, int len)
{
    char *s;
    unsigned char *t;
    int i;
    /*
       printf("wc2mb(%d)", len);
     */
    t = (unsigned char *) text;
    if (len * 2 > extent_buf_size) {
	if (extent_buf)
	    XFree(extent_buf);
	extent_buf = malloc(len * 2);
	extent_buf_size = len * 2;
    }
    s = extent_buf;
    for (i = 0; i < len * sizeof(wchar_t); i++) {
	if (t[i])
	    *s++ = t[i];
    }
    return (s - extent_buf);
}
#endif

#if WC_DEFINE
int XwcTextExtents(XFontSet fontset, const wchar_t * text, int len, XRectangle * ink, XRectangle * log)
{
    int n;
    n = wc2mb(text, len);
    return XmbTextExtents(fontset, (_Xconst char *) extent_buf, n, ink, log);
}
#endif

int XmbTextExtents(XFontSet fontset, _Xconst char *text, int len, XRectangle * ink, XRectangle * logical)
{
    int width;
    int i;
    XFontStruct *xfont, *hfont;
    int direction;
    XCharStruct overall;
    XCharStruct tmp_overall;
    int tmp_ascent, tmp_descent;
    int logical_ascent, logical_descent;
    int hangul_ascent, hangul_descent, hangul_width;

    logical_ascent = -3000;
    logical_descent = -3000;
    overall.lbearing = 3000;
    overall.rbearing = -1000;
    overall.ascent = -3000;
    overall.descent = -3000;
    width = 0;

    xfont = ENG_FONT(fontset);
    hfont = H_FONT(fontset);

    hangul_ascent = hfont->ascent;
    hangul_descent = hfont->descent;
    hangul_width = H_FONT_WIDTH(fontset);

    while (len > 0) {
	if (*text & 0x80) {
	    for (i = 0; len > 0 && (*text & 0x80); text += 2, len -= 2, i += 2);
	    tmp_ascent = hangul_ascent;
	    tmp_descent = hangul_descent;
	    tmp_overall.lbearing = 0;
	    tmp_overall.rbearing = hangul_width * i / 2;
	    tmp_overall.ascent = tmp_ascent;
	    tmp_overall.descent = tmp_descent;
	    tmp_overall.width = tmp_overall.rbearing;
	    width += hangul_width * i / 2;
	} else {
	    for (i = 0; len > 0 && !(*text & 0x80); text++, len--, i++);
	    XTextExtents(xfont, text - i, i,
		    &direction, &tmp_ascent, &tmp_descent, &tmp_overall);
	    width += tmp_overall.rbearing;
	}
	overall.lbearing = min(overall.lbearing,
			       overall.width + tmp_overall.lbearing);
	overall.rbearing = max(overall.rbearing,
			       overall.width + tmp_overall.rbearing);
	overall.ascent = max(overall.ascent, tmp_overall.ascent);
	overall.descent = max(overall.descent, tmp_overall.descent);
	overall.width += tmp_overall.width;
	logical_ascent = max(logical_ascent, tmp_ascent);
	logical_descent = max(logical_descent, tmp_descent);
    }

    if (ink) {
	ink->x = overall.lbearing;
	ink->y = -overall.ascent;
	ink->width = overall.rbearing - overall.lbearing;
	ink->height = overall.ascent + overall.descent;
    }
    if (logical) {
	logical->width = width;
	logical->height = logical_descent + logical_ascent;
	logical->x = 0;
	logical->y = -logical_ascent;
    }
#if DEBUG
    printf("exwidth: %d\n", width);
#endif
    return width;
}


#if WC_DEFINE
Status
XwcTextPerCharExtents(XFontSet font_set, xconst wchar_t * text,
		 int len, XRectangle * ink_buf, XRectangle * logical_buf,
		      int buf_size, int *num_chars,
		  XRectangle * overall_ink, XRectangle * overall_logical)
{
    int n;
#if DEBUG
    printf("%s\n", __FUNCTION__);
#endif
    n = wc2mb(text, len);
    return XmbTextPerCharExtents(font_set, (_Xconst char *) extent_buf, n, ink_buf,
	 logical_buf, buf_size, num_chars, overall_ink, overall_logical);
}
#endif

Status
XmbTextPerCharExtents(XFontSet font_set, _Xconst char *text,
		 int len, XRectangle * ink_buf, XRectangle * logical_buf,
		      int buf_size, int *num_chars,
		  XRectangle * overall_ink, XRectangle * overall_logical)
{
    XFontStruct *font;
    Bool is_hangul;
    XCharStruct *def, *cs, overall;
    XCharStruct hangul_cs;
    int buf_len, require_num;
    int logical_ascent, logical_descent;
    Bool first = True;
    int i;


    memset((char *) &overall, 0, sizeof(XCharStruct));
    logical_ascent = logical_descent = require_num = *num_chars = 0;
    hangul_cs.lbearing = 0;
    hangul_cs.rbearing = H_FONT_WIDTH(font_set);
    hangul_cs.ascent = H_FONT(font_set)->ascent;
    hangul_cs.descent = H_FONT(font_set)->descent;
    hangul_cs.width = H_FONT_WIDTH(font_set);

    while (len > 0) {

	if (*text & 0x80) {
	    for (i = 0; i < len && (text[i] & 0x80); i += 2);
	    buf_len = i;
	    is_hangul = 1;
	    font = H_FONT(font_set);
	} else {
	    for (i = 0; i < len && !(text[i] & 0x80); i++);
	    buf_len = i;
	    is_hangul = 0;
	    font = ENG_FONT(font_set);
	}

	if (require_num) {
	    require_num += buf_len;
	    continue;
	}
	if (buf_size < buf_len) {
	    require_num = *num_chars + buf_len;
	    continue;
	}
	buf_size -= buf_len;

	if (first) {
	    logical_ascent = font->ascent;
	    logical_descent = font->descent;
	} else {
	    logical_ascent = max(logical_ascent, font->ascent);
	    logical_descent = max(logical_descent, font->descent);
	}

	if (is_hangul) {
	} else {
	    CI_GET_DEFAULT_INFO_1D(font, def);
	}

	while (buf_len-- > 0) {
	    if (is_hangul) {
		text += 2;
		cs = &hangul_cs;
	    } else {
		CI_GET_CHAR_INFO_1D(font, *text, def, cs);
		text++;
	    }
	    if (cs == NULL)
		continue;

	    ink_buf->x = overall.width + cs->lbearing;
	    ink_buf->y = -(cs->ascent);
	    ink_buf->width = cs->rbearing - cs->lbearing;
	    ink_buf->height = cs->ascent + cs->descent;
	    ink_buf++;

	    logical_buf->x = overall.width;
	    logical_buf->y = -(font->ascent);
	    logical_buf->width = cs->width;
	    logical_buf->height = font->ascent + font->descent;
	    logical_buf++;

	    if (first) {
		overall = *cs;
		first = False;
	    } else {
		overall.ascent = max(overall.ascent, cs->ascent);
		overall.descent = max(overall.descent, cs->descent);
		overall.lbearing = min(overall.lbearing,
				       overall.width + cs->lbearing);
		overall.rbearing = max(overall.rbearing,
				       overall.width + cs->rbearing);
		overall.width += cs->width;
	    }

	    (*num_chars)++;
	}
    }

    if (require_num) {
	*num_chars = require_num;
	return 0;
    } else {
	if (overall_ink) {
	    overall_ink->x = overall.lbearing;
	    overall_ink->y = -(overall.ascent);
	    overall_ink->width = overall.rbearing - overall.lbearing;
	    overall_ink->height = overall.ascent + overall.descent;
	}
	if (overall_logical) {
	    overall_logical->x = 0;
	    overall_logical->y = -(logical_ascent);
	    overall_logical->width = overall.width;
	    overall_logical->height = logical_ascent + logical_descent;
	}
    }

    return 1;

}

void XmbDrawText(Display * display, Drawable window,
		 GC gc, int x, int y, XmbTextItem * text_items, int nitem)
{
    XFontSet fs;
    XmbTextItem *p = text_items;
    int i = nitem;

    while (i && !p->font_set) {
	i--;
	p++;
    }

    for (; --i >= 0; p++) {
	if (p->font_set)
	    fs = p->font_set;
	x += p->delta;
	XmbDrawString(display, window, fs, gc, x, y, p->chars, p->nchars);
	x += XmbTextEscapement(fs, p->chars, p->nchars);
    }
}

static int default_draw(Display * display, Drawable drawable, GC gc, int x, int y,
			_Xconst char *text, int len, XFontSet fontset)
{
    XFontStruct *xfont;
#if DEBUG
    int i;
    for (i = 0; i < len; i++)
	putchar(text[i]);
    putchar('\n');
#endif
    xfont = ENG_FONT(fontset);
    XSetFont(display, gc, xfont->fid);
    XDrawString(display, drawable, gc, x, y, text, len);
    return x + XTextWidth(xfont, text, len);
}

static XChar2b *draw_buf = NULL;
static int draw_buf_len = 0;

static int ksc1_draw(Display * display, Drawable drawable, GC gc, int x, int y,
		     _Xconst char *text, int len, XFontSet fontset)
{
    XChar2b *buf;
    int i;

    if (len > draw_buf_len) {
	if (len)
	    XFree(draw_buf);
	draw_buf = (XChar2b *) malloc(sizeof(XChar2b) * len);
	draw_buf_len = len;
    }
    buf = draw_buf;
    i = 0;
    while (i++ < len) {
	buf->byte1 = *text++;
	buf->byte2 = *text++;
	buf++;
    }

    XSetFont(display, gc, H_FONT(fontset)->fid);
    XDrawString16(display, drawable, gc, x, y, draw_buf, len);

    return x + H_FONT_WIDTH(fontset) * len;
}

static int ksc0_draw(Display * display, Drawable drawable, GC gc, int x, int y,
		     _Xconst char *text, int len, XFontSet fontset)
{
    XChar2b *buf;
    int i;

#if DEBUG
    printf("%d %d gc: %p fontset: %p\n", x, y, gc, fontset);
#endif
    if (len > draw_buf_len) {
	if (len)
	    XFree(draw_buf);
	draw_buf = (XChar2b *) malloc(sizeof(XChar2b) * len);
	draw_buf_len = len;
    }
    buf = draw_buf;

    i = 0;
    while (i++ < len) {
	buf->byte1 = (*text++) & 0x7F;
	buf->byte2 = (*text++) & 0x7F;
	buf++;
    }

    XSetFont(display, gc, H_FONT(fontset)->fid);
    XDrawString16(display, drawable, gc, x, y, draw_buf, len);

    return x + H_FONT_WIDTH(fontset) * len;
}

static int johab_draw(Display * display, Drawable drawable, GC gc, int x, int y,
		      _Xconst char *text, int len, XFontSet fontset)
{
    int n;

    if (len * 3 > draw_buf_len) {
	if (len)
	    XFree(draw_buf);
	draw_buf = (XChar2b *) malloc(sizeof(XChar2b) * len * 3);
	draw_buf_len = len * 3;
    }
    n = convert_ks_to_johabfont(text, draw_buf, len);
    XSetFont(display, gc, H_FONT(fontset)->fid);
    XDrawString16(display, drawable, gc, x, y, draw_buf, n);

    return x + H_FONT_WIDTH(fontset) * len;
}

static int johab844_draw(Display * display, Drawable drawable, GC gc, int x, int y,
			 _Xconst char *text, int len, XFontSet fontset)
{
    int n;

    if (len * 3 > draw_buf_len) {
	if (len)
	    XFree(draw_buf);
	draw_buf = (XChar2b *) malloc(sizeof(XChar2b) * len * 3);
	draw_buf_len = len * 3;
    }
    n = convert_ks_to_johab844font(text, draw_buf, len);
    XSetFont(display, gc, H_FONT(fontset)->fid);
    XDrawString16(display, drawable, gc, x, y, draw_buf, n);

    return x + H_FONT_WIDTH(fontset) * len;
}
