
/* this file is a part of gau software, (C) Hwang chi-duck 1997               */

#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>


#include "gau.h"
#include "config.h"
#include <gdk/gdkkeysyms.h>

int srz_pid = -1;
int use_internal_rz = TRUE;
gboolean exit_after_download = FALSE;
static void run_real_sz(char *filename);

int
zmodem_input_filter(char *str, int bytes)
{
	static char           zmSig[] = "**\030B00";
	static char          *zmSigPtr = zmSig;
	int i;
	/* check signature of zmodem */
	for(i=0;i<bytes;i++) {
		if(str[i] != *zmSigPtr) zmSigPtr = zmSig;
		else if(*++zmSigPtr == '\0') {
			run_rz();
			return TRUE;
		}
	}
	return FALSE;
}

void rz_signal_child_handle(int sig);
void sz_signal_child_handle(int sig);

int srz_input_tag;
int srzpipefd[2];

void 
srz_input_handler(gpointer data, int source, GdkInputCondition con)
{
	char buf[256];
	int size = read(source, buf, 255);
	term->to_term(term, buf, size);
}


void
run_rz(void)
{
	gau_state = DOWNLOAD;
	gtk_widget_set_sensitive(hangup_after_transfer_button, TRUE);
	if(use_internal_rz) {
		run_internal_rz();
		return;
	}
	pipe(srzpipefd);
	srz_pid = fork();
	if(srz_pid > 0) {
		/* parent */
		signal(SIGCHLD, rz_signal_child_handle);
		gdk_input_remove(read_tag);
		close(srzpipefd[1]);
		srz_input_tag = 
			gdk_input_add(srzpipefd[0], GDK_INPUT_READ, 
				srz_input_handler, NULL);
		return;
	}
	chdir(expand_path(DOWN_PATH));
	dup2(read_fd, 0);
	dup2(read_fd, 1);
	close(read_fd);
	close(srzpipefd[0]);
	dup2(srzpipefd[1], 2);
	execlp("lrz", "rzgrim_show", "-r", "-vv", NULL);
	g_warning("%s : fail to run rz\n", __FUNCTION__);
	exit(1);
}

static void
key_press_callback(GtkWidget *w, GdkEventKey *ev, gpointer data)
{
	if(ev->keyval == GDK_Return) {
		char *s = gtk_entry_get_text(GTK_ENTRY(w));
		run_real_sz(s);
		gtk_widget_destroy(GTK_WIDGET(data));
	}
}

void
run_sz(void)
{
	GtkWidget *win;
	GtkWidget *vbox, *entry, *button, *label, *hbox;
	gtk_widget_set_sensitive(hangup_after_transfer_button, TRUE);
	win = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(win), " ");
	gtk_window_position(GTK_WINDOW(win), GTK_WIN_POS_CENTER);
	vbox = gtk_vbox_new(FALSE, 2);
	gtk_container_add(GTK_CONTAINER(win), vbox);
	gtk_widget_show(vbox);

	hbox = gtk_hbox_new(FALSE, 3);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);
	gtk_widget_show(hbox);

	label = gtk_label_new(" ̸:");
	gtk_container_add(GTK_CONTAINER(hbox), label);
	gtk_widget_show(label);

	entry = gtk_entry_new();
	gtk_container_add(GTK_CONTAINER(hbox), entry);
	gtk_signal_connect(GTK_OBJECT(entry), "key_press_event",
		(GtkSignalFunc)key_press_callback, win);
	gtk_widget_show(entry);
	gtk_widget_new(gtk_hseparator_get_type(),
		"GtkWidget::parent", vbox,
		"GtkWidget::visible", TRUE,
		NULL);
	button = gtk_button_new_with_label("");
	gtk_container_add(GTK_CONTAINER(vbox), button);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(win));
	gtk_widget_show(button);
	gtk_widget_show(win);
}


static void
run_real_sz(char *filename)
{
	struct stat buf;
	if(filename == NULL) {
		g_warning("in %s filename is NULL!!!\n", __FUNCTION__);
		return;
	}
	if(stat(filename, &buf) != 0) {
		g_warning("In \"%s\", fail to stat '%s'\n", __FUNCTION__, filename);
		return;
	}
	pipe(srzpipefd);
	srz_pid = fork();
	if(srz_pid > 0) {
		/* parent */
		signal(SIGCHLD, sz_signal_child_handle);
		gdk_input_remove(read_tag);
		close(srzpipefd[1]);
		srz_input_tag = gdk_input_add(srzpipefd[0], GDK_INPUT_READ, srz_input_handler, NULL);
		gau_state = UPLOAD;
		return;
	}
	// child
	dup2(read_fd, 0);
	dup2(read_fd, 1);
	close(read_fd);
	dup2(srzpipefd[1], 2);
	close(srzpipefd[0]);
	execlp("lsz", "lsz", filename, NULL);
	g_warning("%s : fail to run sz\n", __FUNCTION__);
	exit(1);
}

void
rz_cancel(void)
{
	if(srz_pid <= 0) return;
	kill(srz_pid, SIGTERM);
}

void
sz_cancel(void)
{
	if(srz_pid <= 0) return;
	kill(srz_pid, SIGTERM);
}

void
rz_signal_child_handle(int sig)
{
	int status;
	if(wait(&status) != srz_pid) {
		signal(SIGCHLD, rz_signal_child_handle);
		return;
	}
	if(!WIFEXITED(status) || (WIFEXITED(status) && WEXITSTATUS(status) != 0)) {
		// killed by a signal which was not caught;
		// or exit code is not zero;
		static char canistr[] = {
		 24,24,24,24,24,24,24,24,24,24,8,8,8,8,8,8,8,8,8,8,0
		};
		g_print("rz exit abnormally\n");
		term->from_term(canistr, strlen(canistr));
	} 

	if(use_modem) signal(SIGCHLD, SIG_IGN);
	else if(!use_internal_telnet) {
		signal(SIGCHLD, telnet_child_signal);
	} else { // using internal telnet
		signal(SIGCHLD, SIG_IGN);
	}
	if(exit_after_download) {
		hangup();
	} 
	gtk_widget_set_sensitive(hangup_after_transfer_button, FALSE);
	play_transfer_end();
	srz_pid = -1;
	gdk_input_remove(srz_input_tag);
	close(srzpipefd[0]);
	read_tag = gdk_input_add(read_fd, GDK_INPUT_READ, 
		input_handler_from_external, NULL);
}

void
sz_signal_child_handle(int sig)
{
	if(wait(NULL) != srz_pid) {
		signal(SIGCHLD, sz_signal_child_handle);
		return;
	}
	gdk_input_remove(srz_input_tag);
	read_tag = gdk_input_add(read_fd, GDK_INPUT_READ, 
		input_handler_from_external, NULL);
	if(exit_after_download) {
		hangup();
	} 
	gtk_widget_set_sensitive(hangup_after_transfer_button, FALSE);
	play_transfer_end();
	srz_pid = -1;
}

int zmodem_input_tag;

void
run_internal_rz(void)
{
	zmodem_init();
	gdk_input_remove(read_tag);
	zmodem_input_tag = gdk_input_add(read_fd, 
		GDK_INPUT_READ, zmodem_input_handler, NULL);
}

void
end_internal_rz(void)
{
	gdk_input_remove(zmodem_input_tag);
	read_tag = gdk_input_add(read_fd, GDK_INPUT_READ, 
		input_handler_from_external, NULL);
	play_transfer_end();
	gtk_widget_set_sensitive(hangup_after_transfer_button, FALSE);
}

